part of TeamSpeak_UserTrack;

class UserTracker {
  static const int BandwidthTalkThreshold = 400;
  
  /// Server where users should be tracked on
  final Server server;
  
  /// Sets whether events should be fired for serverquery-clients as well.
  /// By default, only "normal" users are tracked.
  final bool trackServerQueryClients;

  final StreamController<UserEvent> _controller = new StreamController();

  /// Stream that provides all tracking events (no broadcast!)
  Stream<UserEvent> get stream => _controller.stream;
  
  void _addEvent(UserEvent event){
    if(!(event.user.is_serverquery_client && !trackServerQueryClients))
      _controller.add(event);
  }
  
  void _enterServer(User user){
    final UserEnteredServerEvent entered = new UserEnteredServerEvent(user);
    _lastEvents[user.database_id] = [entered];
    _addEvent(entered);
  }
  
  // ClientDatabaseIds used as values
  Map<int, List<UserEvent>> _lastEvents = new Map();
  
  void _checkTalk(UserStoppedTalkingEvent event){
    // Check if we just switched channel and if length is just one
    bool ignore = false;
    if(event.talkTime.inMilliseconds.abs() < 2000){
      // Get last channel enter event
      final UserEnteredChannelEvent entered = _lastEvents[event.user.database_id]
        .singleWhere((e) => e is UserEnteredChannelEvent);
      if(entered.time.difference(new DateTime.now()).inMilliseconds.abs() < 2000){
        print("Ignoring...");
        ignore = true;
      }
    }
    if(!ignore) _addEvent(new UserTalkedEvent(event.user, event.time, event.talkTime));
  }
  
  final List<StreamSubscription> _subscriptions = new List();
  
  /// Ends the tracker and cancels all update subscriptions
  void close(){
    _subscriptions.forEach((e) => e.cancel());
  }
  
  /*
   * Initializes a new UserTracker which is bound to an already initialized
   * teamspeak client. The Tracker then listens on the event streams of the
   * client and gathers information about it.
   */
  UserTracker(this.server, {this.trackServerQueryClients: false}){
    if(server == null)
      throw new StateError("The TeamSpeak client is not initialized yet.");
    if(server.enableThrottling)
      throw new StateError("UserTracker can not be used on throttled connections. Bandwidth usage would be too high.");
    if(server.autoUpdate >= new Duration(seconds: 2))
      throw new StateError("UserTracker should be used on a connection that is auto updated every second.");
  
    final Map<int, int> userChannelCache = new Map();
    
    final enterSubscription = server.onUserEnter.listen((UserEnterServerEvent event)
        => _enterServer(event.user));
    
    final leaveSubscription = server.onUserLeave.listen((UserLeaveServerEvent event){
      final UserEnteredServerEvent entered = _lastEvents[event.user.database_id]
        .singleWhere((e) => e is UserEnteredServerEvent);
      final UserEnteredChannelEvent enteredChannel = _lastEvents[event.user.database_id]
        .singleWhere((e) => e is UserEnteredChannelEvent);
      
      _addEvent(new UserLeftChannelEvent(event.user, enteredChannel.channel, entered.time, entered.time.difference(new DateTime.now())));
      _addEvent(new UserLeftServerEvent(event.user, entered.time, entered.time.difference(new DateTime.now())));
      
      final Iterable<UserStartedTalkingEvent> talkings = _lastEvents[event.user.database_id]
        .where((e) => e is UserStartedTalkingEvent);
      if(talkings.length > 0){
        final UserStoppedTalkingEvent stoptalking = new UserStoppedTalkingEvent(event.user, talkings.single.time, 
            talkings.single.time.difference(new DateTime.now()));
        _addEvent(stoptalking);
        _lastEvents[event.user.database_id] = _lastEvents[event.user.database_id]
          .where((e) => !(e is UserStartedTalkingEvent)).toList(growable: true);
        _checkTalk(stoptalking);
      }
      
      _lastEvents.remove(event.user.database_id);
    });
    
    server.update().then((_){
      // Use first clients list to fire enter events for users that are alread logged in
      server.users.values.forEach((User user){
        _enterServer(user);
      });
      final updateSubscription = server.onUpdate.listen((_){
        final List<int> onlineUsers = userChannelCache.values.toList(growable: false);
        
        // Check if users changed their channels
        server.users.values.forEach((user){
          // Events are not setup yet. Wtf?
          if(_lastEvents[user.database_id] == null){
            return;
          }
          
          if(userChannelCache[user.database_id] != user.cid){
            // First, send channel left event for the last channel (if any)
            if(userChannelCache[user.database_id] != null){
              // Get last enter event to calculate timespan
              final UserEnteredChannelEvent entered = _lastEvents[user.database_id]
                .singleWhere((e) => e is UserEnteredChannelEvent);
              
              // Delete last UserLeftChannel and UserEnteredChannel events
              _lastEvents[user.database_id] = _lastEvents[user.database_id].
                  where((e) => !(e is UserLeftChannelEvent) || (e is UserEnteredChannelEvent)).toList(growable: true);
              final UserLeftChannelEvent left = new UserLeftChannelEvent(user,
                  (server.channels.containsKey(userChannelCache[user.database_id])
                      ? server.channels[userChannelCache[user.database_id]]
                      : null),
                  entered.time,
                  entered.time.difference(new DateTime.now()));
              _lastEvents[user.database_id].add(left);
              _addEvent(left);
              
              // Check if user was talking and send UserStoppedTalkingEvent if neccessary
              if(_lastEvents[user.database_id]
                .where((e) => e is UserStartedTalkingEvent).length > 0){
                final UserStartedTalkingEvent talking = _lastEvents[user.database_id]
                  .singleWhere((e) => e is UserStartedTalkingEvent);
                final UserStoppedTalkingEvent stoptalking = new UserStoppedTalkingEvent(user, talking.time, 
                    talking.time.difference(new DateTime.now()));
                _addEvent(stoptalking);
                _lastEvents[user.database_id] = _lastEvents[user.database_id]
                  .where((e) => !(e is UserStartedTalkingEvent)).toList(growable: true);
              }
            }
            
            // Then, send channel joined event for the new channel
            // Delete last UserEnteredChannel event
            _lastEvents[user.database_id] = _lastEvents[user.database_id].
                where((e) => !(e is UserEnteredChannelEvent)).toList(growable: true);
            // Send new UserEnteredChannel event
            final UserEnteredChannelEvent entered = new UserEnteredChannelEvent(user, user.channel);
            _lastEvents[user.database_id].add(entered);
            _addEvent(entered);
            userChannelCache[user.database_id] = user.channel.id;
          }
          
          // User started talking
          if(user.bandwidth_received_last_second_total >= BandwidthTalkThreshold){
            if(_lastEvents[user.database_id]
              .where((e) => e is UserStartedTalkingEvent).length == 0){
              
              // User just started talking
              final UserStartedTalkingEvent talkingevent = new UserStartedTalkingEvent(user);
              _addEvent(talkingevent);
              _lastEvents[user.database_id].add(talkingevent);
            }
          } else if(_lastEvents[user.database_id]
              .where((e) => e is UserStartedTalkingEvent).length > 0) {
            
            // User stopped talking!
            final UserStartedTalkingEvent talking = _lastEvents[user.database_id]
              .singleWhere((e) => e is UserStartedTalkingEvent);
            final UserStoppedTalkingEvent stoptalking = new UserStoppedTalkingEvent(user, talking.time, 
                  talking.time.difference(new DateTime.now()));
            _addEvent(stoptalking);
            
            // Strip last talkingstart events from the event list so the client
            // no longer is considered "talking"
            _lastEvents[user.database_id] = _lastEvents[user.database_id]
              .where((e) => !(e is UserStartedTalkingEvent)).toList(growable: true);
            _checkTalk(stoptalking);
          }
          
        });
      });
      
      // Add stream subscriptions so that they can be closed later
      _subscriptions.addAll([enterSubscription,
                             leaveSubscription,
                             updateSubscription]);
    });
  }
}